# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
More detailed representations of data from the nodes.
"""

from ironflow.gui.boxes.node_interface.control import NodeController
from ironflow.gui.boxes.node_interface.representation import NodePresenter
